#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <time.h>
#include <fitsio.h>
#include <math.h>
#include "pvo-FITS.hh"
#include "pvo-utils.hh"
#include "pvo-block.hh"
#include "pvo-header.hh"

using namespace PVO::Block;
using namespace PVO::Header;

namespace PVO
{

  namespace FITS
  {

    void PSRFits::fixPSRFITS()
    {
      /**
       *  Corrects known errors that exist in the PSRFITS definition:
       *  <ul><li> Set the TFORM keyword for the data column in the
       *             SUBINT table
       *      <li> Remove the unwanted Binary Tables.
       *  </ul>
       */
      if(this->gotoSubint()){
	PVOError("PSRFits::fixPSRFITS","Could not go to the SUBINT table.\n");
	return;
      }
    
      int status=0;
      char cval[9];
      strcpy(cval,"16X");
      fits_update_key(this->fptr, TSTRING, "TFORM19", cval, NULL, &status);
      if(status) this->FITSError("PSRFits::fixPSRFITS 1", status);

      if((status=this->deleteTable("BANDPASS")))
	this->FITSError("PSRFits::deleteTable(BANDPASS)", status);
      if((status=this->deleteTable("COHDDISP")))
	this->FITSError("PSRFits::deleteTable(COHDDISP)", status);
      if((status=this->deleteTable("PSREPHEM")))
	this->FITSError("PSRFits::deleteTable(PSREPHEM)", status);
      if((status=this->deleteTable("POLYCO")))
	this->FITSError("PSRFits::deleteTable(POLYCO)", status);
      if((status=this->deleteTable("T2PREDICT")))
	this->FITSError("PSRFits::deleteTable(T2PREDICT)", status);
      if((status=this->deleteTable("FLUX_CAL")))
	this->FITSError("PSRFits::deleteTable(FLUX_CAL)", status);
      if((status=this->deleteTable("CAL_POLN")))
	this->FITSError("PSRFits::deleteTable(CAL_POLN)", status);
      if((status=this->deleteTable("FEEDPAR")))
	this->FITSError("PSRFits::deleteTable(FEEDPAR)", status);
      if((status=this->deleteTable("DIG_STAT")))
	this->FITSError("PSRFits::deleteTable(DIG_STAT)", status);
      if((status=this->deleteTable("DIG_CNTS")))
	this->FITSError("PSRFits::deleteTable(DIG_CNTS)", status);

      this->gotoPHU();
    }

    //--------------------------------------------------------------------------

    int PSRFits::initialise(DataBlock &block)
    {
      /**
       * Initialises a fits file for use in the conversion program. The
       * order of business is as follows: 
       *
       * <ul>
       *   <li> When new:
       *   <ul><li> Set the initial date & time (UT) for the file. 
       *       <li> If a new file is beginning, and the time has changed
       *       and is smaller than the existing time, we need to
       *       increment the date. This may be the case if we are
       *       working through a large tape file and the blocks go past
       *       midnight. 
       *   </ul>
       *   <li> Set the beam used in the file.
       *   <li> Define the FITS file name.
       *   <li> Set the starting block value.
       *   <li> If new, test whether file exists.
       *   <li> Open the fits file.
       *   <li> If new:
       *   <ul><li> Write the PHU headers using writePHU();
       *       <li> Setup and write the SUBINT header with setupSubint(). 
       *   </ul>
       * </ul>
       * \param block The block used to get the header information.
       */

      std::string newUT,date;

      if(this->newFile){

	this->numSamples = 0; // Initialise the number of samples.

	newUT = block.getKeyword<std::string>(UT);
	date   = block.getKeyword<std::string>(DATE);

	// if we have gone over midnight.
	if(dmsToDec(this->UT_initial) > dmsToDec(newUT) ) 
	  date = addDay(date);

	this->setUT(newUT);
	this->setDate(date);
      }

      this->setBeam(block.getKeyword<int>(IBEAM));

      if(block.getSurveyType()==S70) this->setInitialBlock(block.getKeyword<int>(BLK_CNTR));
      else this->setInitialBlock(1);

      this->setFITSname(block.getSurveyType());

      if(this->newFile && this->fileExists()){
	std::stringstream errmsg;
	errmsg << "\aThe file "<< this->getFITSname() 
	       << " already exists.\n"
	       << "Don't want to overwrite, so exiting...\n";
	if(this->verbose) PVOError("FITS initialise",errmsg.str());
	return FAILURE;
      }

      if(this->openFile()) {
	if(this->verbose) 
	  std::cerr << "\aCannot open file "<< this->getFITSname()
		    << ".\nExiting...\n";
	return FAILURE;
      }

      if(this->newFile){
      
	//      this->writeHistoryTapeFile();
	this->writeHistoryCommand(block);
  
	this->writePHU(block);

	this->setupSubint(block);
    
      }

      this->newFile = false;

      return SUCCESS;

    }
    //--------------------------------------------------------------------------

    void PSRFits::setupSubint(DataBlock &block)
    {
      /**
       *  A function to set up the subintegration binary table
       *  extension.  This defines the necessary keywords, as well as
       *  the DAT_... arrays to go in the columns (as these will be the
       *  same for all rows of the subintegration table -- at least as
       *  far as the conversion program goes.)
       *
       *  It then calls the Subint::writeKeywords() function to write
       *  the header keywords to the table.
       *
       * \param block The block containing the header information.
       */
    
      this->subintTable.tsamp = block.getKeyword<float>(SAMP_INT,0)/1000.;
      this->subintTable.nbits = block.getKeyword<int>(NBITS,0);
      this->subintTable.nbin = 1; 
      this->subintTable.nch_file = block.getKeyword<int>(NCHAN);
      this->subintTable.npol = 1;
      this->subintTable.nsblk = block.getKeyword<int>(SAMP_GRP,0) * 
	block.getKeyword<int>(GRP_BLK);

      // define the DAT_FREQ array;
      float deltaFreq = block.getKeyword<float>(CHANBW);
      int numChan = this->subintTable.nch_file;
      this->subintTable.dat_freq = new float[numChan];
      this->subintTable.dat_freq[0] = block.getKeyword<float>(FREQ_CHAN1);
      for(int i=1;i<this->subintTable.nch_file;i++) 
	this->subintTable.dat_freq[i] = 
	  this->subintTable.dat_freq[i-1] + deltaFreq;
/*
// This is to switch the order of the frequencies
      this->subintTable.dat_freq[numChan-1] = block.getKeyword<float>(FREQ_CHAN1);
      for(int i=numChan-1;i>0;i--) 
	this->subintTable.dat_freq[i-1] = 
	  this->subintTable.dat_freq[i] + deltaFreq;
*/

      // define the DAT_WTS array;
      this->subintTable.dat_wts = new float[this->subintTable.nch_file];
      for(int i=0;i<this->subintTable.nch_file;i++)
	this->subintTable.dat_wts[i] = 1.;

      // define the DAT_OFFS array;
      this->subintTable.dat_offs = new float[this->subintTable.nch_file];
      for(int i=0;i<this->subintTable.nch_file;i++)
	this->subintTable.dat_offs[i] = 0.;

      // define the DAT_SCL array;
      this->subintTable.dat_scl = new float[this->subintTable.nch_file];
      for(int i=0;i<this->subintTable.nch_file;i++)
	this->subintTable.dat_scl[i] = 1.;
    
    
      this->subintTable.setAllocatedFlag(true);

      // now write the keywords to the subint table
      if(this->gotoSubint()) 
	PVOError("PSRFits::initialise","Could not go to the SUBINT table.\n");
      else{
	int status = this->subintTable.writeKeywords(this->fptr);
	if(status) this->FITSError("Subint::writeKeywords",status);
      }

    
    }

    //-----------------------------------------------------------------------

    int Subint::writeKeywords(fitsfile *fptr)
    {
      /**
       *  A special function in the Subint class to write all the header
       *  keywords necessary for the subintegration binary table
       *  extension. This should only need to be done once per file.
       *
       *  The keywords are stored as private members of Subint, and need
       *  to be defined beforehand.
       *
       *  \param fptr The fitsfile pointer corresponding to the file to
       *  be written to.
       */

      int status=0;
      fits_update_key(fptr,TSTRING,"INT_TYPE", 
		      (char *)this->int_type.c_str(), NULL, &status);
      if(status) return status;
      else{
	
	status=0;
	fits_update_key(fptr,TSTRING,"INT_UNIT",
			(char *)this->int_unit.c_str(),NULL,&status);
	status=0;
	fits_update_key(fptr,TFLOAT,"TSAMP",&this->tsamp,NULL,&status);    
	status=0;
	fits_update_key(fptr,TINT,"NBITS",&this->nbits,NULL,&status);    
	status=0;
	fits_update_key(fptr,TINT,"NBIN",&this->nbin,NULL,&status);
	status=0;
	fits_update_key(fptr,TINT,"NCH_FILE",&this->nch_file,NULL,&status);
	status=0;
	fits_update_key(fptr,TINT,"NCH_STRT",&this->nch_strt,NULL,&status);
	status=0;
	fits_update_key(fptr,TINT,"NPOL",&this->npol,NULL,&status);
	status=0;
	fits_update_key(fptr,TINT,"NSBLK",&this->nsblk,NULL,&status);
	status=0;
	int colnum=0;
	fits_get_colnum(fptr, CASEINSEN, "DATA", &colnum, &status);
	status=0;
	fits_modify_vector_len (fptr, colnum, this->getDataSize(), &status);

	// delete the existing "TDIM19 = (*,*,*,*)" and write the new one.
	status=0;
	fits_delete_key(fptr, "TDIM19", &status);
	int naxis=4;
	long naxes[4];
	naxes[0] = this->nbits;
	naxes[1] = this->nch_file;
	naxes[2] = this->npol;
	naxes[3] = this->nsblk;
	status=0;
	fits_write_tdim(fptr, colnum, naxis, naxes, &status);

      }
      return 0;
    }
    //--------------------------------------------------------------------------

    void PSRFits::writeSubintRow(DataBlock &block)
    {
      /**
       * This function writes an entire new row in the SUBINT table extension.
       * The procedure is as follows:
       * <ul> <li> Move to the subint table.
       *      <li> Start a new row with startNewSubintRow()
       *      <li> Write the single-valued columns with writeColumnValue()
       *      <li> Write the multi-valued DAT_... columns
       *      <li> Write the data.
       *      <li> Increment the number of samples.
       * </ul>
       * \param block The block containing the header info & the data.
       */

      this->gotoSubint();
      this->startNewSubintRow(); 

      float tsubint = block.getKeyword<float>(BLK_SEC);
      double off_sub;
    
      off_sub = 0.5*(2.*(block.getKeyword<int>(BLK_CNTR)-this->blk_initial)+1)
	* block.getKeyword<double>(BLK_SEC);

      double lst_sub  = dmsToDec(block.getKeyword<std::string>(LST));
      double ra_sub   = dmsToDec(block.getKeyword<std::string>(RA))*15.;
      double dec_sub  = dmsToDec(block.getKeyword<std::string>(DEC));
      double glon_sub = block.getKeyword<double>(GL);
      double glat_sub = block.getKeyword<double>(GB);
      float  fd_ang   = block.getKeyword<float>(FEED_ANGLE);
      float  par_ang  = block.getKeyword<float>(PANGLE);
      float  pos_ang  = remainderf(par_ang + fd_ang, 360.);
      float  tel_az   = block.getKeyword<float>(TEL_AZ);
      float  tel_zen  = block.getKeyword<float>(TEL_ZEN);
    
      this->writeColumnValue("TSUBINT",  tsubint);
      this->writeColumnValue("OFFS_SUB", off_sub);
      this->writeColumnValue("LST_SUB",  lst_sub);
      this->writeColumnValue("RA_SUB",   ra_sub);
      this->writeColumnValue("DEC_SUB",  dec_sub);
      this->writeColumnValue("GLON_SUB", glon_sub);
      this->writeColumnValue("GLAT_SUB", glat_sub);
      this->writeColumnValue("FD_ANG",   fd_ang);
      this->writeColumnValue("POS_ANG",  pos_ang);
      this->writeColumnValue("PAR_ANG",  par_ang);
      this->writeColumnValue("TEL_AZ",   tel_az);
      this->writeColumnValue("TEL_ZEN",  tel_zen);

      /////////////////////////////////////

      long nrows = this->getNumRows();
      int colnum = this->getColnum("DAT_FREQ");
      int status=0;
      fits_modify_vector_len (this->fptr, colnum, 
			      this->subintTable.nch_file, &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);
      status=0;
      fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		     this->subintTable.nch_file, this->subintTable.dat_freq, 
		     &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);

      colnum = this->getColnum("DAT_WTS");
      status=0;
      fits_modify_vector_len (this->fptr, colnum, 
			      this->subintTable.nch_file, &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);
      status=0;
      fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		     this->subintTable.nch_file, this->subintTable.dat_wts, 
		     &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);

      colnum = this->getColnum("DAT_OFFS");
      status=0;
      fits_modify_vector_len (this->fptr, colnum, 
			      this->subintTable.nch_file, &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);
      status=0;
      fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		     this->subintTable.nch_file, this->subintTable.dat_offs, 
		     &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);

      colnum = this->getColnum("DAT_SCL");
      status=0;
      fits_modify_vector_len (this->fptr, colnum, 
			      this->subintTable.nch_file, &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);
      status=0;
      fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		     this->subintTable.nch_file, this->subintTable.dat_scl, 
		     &status);
      if(status) this->FITSError("PSRFits::writeSubintRow", status);

      //////////////////////////////////////////

      unsigned char *data = (unsigned char *)block.getData();

      if(this->FITSname == "PM_061222_150342_01.fits" && nrows==1){
	std::cout << charToBinary(data[0]);
	std::cout << charToBinary(data[1]);
	std::cout << charToBinary(data[2]);
	std::cout << "\n";
      }

      colnum = this->getColnum("DATA");
      status=0;
      fits_write_col_byt(this->fptr, colnum, nrows, 1, this->subintTable.getDataSizeBytes(), data, &status);
      if(status) this->FITSError("PSRFits::writeDataArray 4", status);

      //////////////////////////////////////////

      // Increment the number of samples
      this->numSamples += (block.getKeyword<int>(GRP_BLK) * 
			   block.getKeyword<int>(SAMP_GRP) );

    }
    //--------------------------------------------------------------------------

    void PSRFits::writePHU(DataBlock &block)
    {
      /**
       *   Writes the header information contained in the DataBlock block
       *    to the primary header unit.
       *   At the end the current header unit will be the PHU.
       */
      this->gotoPHU();

      char *todaysdate     = new char[25];
      int timeref,status=0;
      fits_get_system_time(todaysdate, &timeref, &status);
      if(status) this->FITSError("PSRFits::writePHU", status);
      std::string projid   = block.getSurveyName();
      std::string telescop = block.getKeyword<std::string>(TELID);
      if((deblank(telescop)=="")&&(block.getSurveyType()==S70)) 
	telescop           = "PARKES";
      double ant_x         = block.Survey().getAntX();
      double ant_y         = block.Survey().getAntY();
      double ant_z         = block.Survey().getAntZ();
      std::string frontend = block.Survey().getFrontend();
      std::string fd_poln  = block.Survey().getFd_poln();
      int fd_hand          = block.Survey().getFd_hand();
      float fd_sang        = block.Survey().getFd_sang();
      float fd_xyph        = block.Survey().getFd_xyph();
      std::string backend  = block.Survey().getBackend();
      std::string beconfig = block.Survey().getBeconfig();
      int be_phase         = block.Survey().getBe_phase();
      int be_dcc           = block.Survey().getBe_dcc();
      double tcycle        = block.Survey().getTcycle();
      int nrcvr            = block.getKeyword<int>(NFILTER);
      std::string obs_mode = block.getObs_mode();

      std::string date_obs = 
	makeDateObs(deblank(block.getKeyword<std::string>(DATE)),
		    deblank(block.getKeyword<std::string>(UT))   );

      double f0            = block.getKeyword<double>(FREQ_CHAN1);
      double bw            = block.getKeyword<double>(CHANBW);
      double nc            = block.getKeyword<double>(NCHAN);
      float obsfreq        = f0 + ((nc-1)/2.) * bw;

      std::string src_name;
      if(block.getSurveyType()==S70) src_name = block.getFieldID();
      else src_name = deblank(block.getKeyword<std::string>(PSRNAME));
      std::string coord_md = block.getCoordMode();
      std::string eq_str   = block.getEquinox();
      float equinox        = 0.;
      if(eq_str==Block::eq_j2000)       equinox = Block::eq_2000;
      else if(eq_str==Block::eq_b1950)  equinox = Block::eq_1950;
      std::string ra       = deblank(block.getKeyword<std::string>(RA));
      std::string dec      = deblank(block.getKeyword<std::string>(DEC));
      float bmaj           = block.Survey().getBmaj();
      float bmin           = block.Survey().getBmin();
      float bpa            = block.Survey().getBpa();
      std::string stt_crd1 = deblank(block.getKeyword<std::string>(RA));
      std::string stt_crd2 = deblank(block.getKeyword<std::string>(DEC));
      std::string trk_mode = block.Survey().getTrk_mode();
      std::string stp_crd1 = deblank(block.getKeyword<std::string>(RA));
      std::string stp_crd2 = deblank(block.getKeyword<std::string>(DEC));
//       float scanlen        = block.Survey().getScanlen();
//      float scanlen        = block.getKeyword<float>(SCAN_LEN);
      std::string fd_mode  = block.Survey().getFd_mode();
      float fa_req         = block.getKeyword<float>(FEED_ANGLE);
      std::string cal_mode = block.Survey().getCal_mode();
      float cal_freq       = block.Survey().getCal_freq();
      float cal_dcyc       = block.Survey().getCal_dcyc();
      float cal_phs        = block.Survey().getCal_phs();
      long stt_imjd        = block.getKeyword<int>(MJD);
      double start_time    = dmsToDec(deblank(block.getKeyword<std::string>(UT)))*3600.;
      long stt_smjd        = long(floor(start_time));
      double stt_offs      = start_time - floor(start_time);
      double stt_lst       = dmsToDec(block.getKeyword<std::string>(LST))*3600.;

      this->updateKeyword("DATE",     todaysdate);
      this->updateKeyword("PROJID",   projid);
      this->updateKeyword("TELESCOP", telescop);
      this->updateKeyword("ANT_X",    ant_x);          
      this->updateKeyword("ANT_Y",    ant_y);        
      this->updateKeyword("ANT_Z",    ant_z);        
      this->updateKeyword("FRONTEND", frontend);
      this->updateKeyword("FD_POLN",  fd_poln);
      this->updateKeyword("FD_HAND",  fd_hand);
      this->updateKeyword("FD_SANG",  fd_sang);
      this->updateKeyword("FD_XYPH",  fd_xyph);
      this->updateKeyword("BACKEND",  backend);
      this->updateKeyword("BECONFIG", beconfig);
      this->updateKeyword("BE_PHASE", be_phase);
      this->updateKeyword("BE_DCC",   be_dcc);
      this->updateKeyword("TCYCLE",   tcycle);
      this->updateKeyword("NRCVR",    nrcvr);
      this->updateKeyword("OBS_MODE", obs_mode);
      this->updateKeyword("DATE-OBS", date_obs);
      this->updateKeyword("OBSFREQ",  obsfreq);
      this->updateKeyword("SRC_NAME", src_name);
      this->updateKeyword("COORD_MD", coord_md);
      if(equinox>0.)
	this->updateKeyword("EQUINOX",  equinox);
      this->updateKeyword("RA",       ra);
      this->updateKeyword("DEC",      dec);
      this->updateKeyword("BMAJ",     bmaj);
      this->updateKeyword("BMIN",     bmin);
      this->updateKeyword("BPA",      bpa);
      this->updateKeyword("STT_CRD1", stt_crd1);
      this->updateKeyword("STT_CRD2", stt_crd2);
      this->updateKeyword("TRK_MODE", trk_mode);
      this->updateKeyword("STP_CRD1", stp_crd1);
      this->updateKeyword("STP_CRD2", stp_crd2);
//       this->updateKeyword("SCANLEN",  scanlen);
      this->updateKeyword("FD_MODE",  fd_mode);
      this->updateKeyword("FA_REQ",   fa_req);
      this->updateKeyword("CAL_MODE", cal_mode);
      this->updateKeyword("CAL_FREQ", cal_freq);
      this->updateKeyword("CAL_DCYC", cal_dcyc);
      this->updateKeyword("CAL_PHS ", cal_phs);
      this->updateKeyword("STT_IMJD", stt_imjd);
      this->updateKeyword("STT_SMJD", stt_smjd);
      this->updateKeyword("STT_OFFS", stt_offs);
      this->updateKeyword("STT_LST ", stt_lst);

      //    this->addNewKeywords(block);

      delete [] todaysdate;

    }
    //-----------------------------------------------------------------------
  
    void PSRFits::addNewKeywords(DataBlock &block)
    {
      /**
       *   Adds new keywords to the PHU that have not yet been added into the 
       *   PSRFITS definition.
       *   Need to be in the PHU before calling this function.
       *
       *   NOTHING LEFT TO DO HERE.
       *  \deprecated
       */
      return;
    }
    //-----------------------------------------------------------------------

    void PSRFits::writeHistoryCommand(DataBlock &block)
    {
      /**
       * Write processing history information to the HISTORY binary table 
       * extension. This is just a row of today's date and what was entered
       * in the command line.
       */
    
      this->gotoHistory();

      char *todaysdate     = new char[25];
      int timeref,status=0;
      fits_get_system_time(todaysdate, &timeref, &status);
      if(status) this->FITSError("PSRFits::writeHistoryCommand", status);

      //    std::string historyMessage;
      //   historyMessage = "PVOCONVERT command: " + this->commandLine;

      this->startNewRow(); 
      this->writeColumnValue("DATE_PRO", todaysdate);
      this->writeColumnValue("PROC_CMD", (char *)this->commandLine.c_str());

      this->writeColumnValue("TBIN", block.getKeyword<double>(SAMP_INT)/1000.);
      double f0            = block.getKeyword<double>(FREQ_CHAN1);
      double bw            = block.getKeyword<double>(CHANBW);
      double nc            = block.getKeyword<double>(NCHAN);
      float obsfreq        = f0 + ((nc-1)/2.) * bw;
      this->writeColumnValue("CTR_FREQ", obsfreq);
      this->writeColumnValue("NCHAN", nc);
      this->writeColumnValue("CHAN_BW", bw);
      

      this->gotoPHU();

      delete [] todaysdate;

    }
    //--------------------------------------------------------------------------

    //   void PSRFits::writeHistoryTapeFile(std::string filename)
    void PSRFits::writeHistoryTapeFile()
    {
      /**
       * Write processing history information to the HISTORY binary table 
       * extension. This is just a row of today's date and what was entered
       * in the command line.
       * /deprecated
       */
    
      this->gotoHistory();

      char *todaysdate     = new char[25];
      int timeref,status=0;
      fits_get_system_time(todaysdate, &timeref, &status);
      if(status) this->FITSError("PSRFits::writeHistoryTapeFile", status);

      std::string historyMessage;
      historyMessage = "Tape file used: " + this->oldFilename;

      this->startNewRow(); 
      this->writeColumnValue("DATE_PRO", todaysdate);
      this->writeColumnValue("PROC_CMD", (char *)historyMessage.c_str());

      this->gotoPHU();

      delete [] todaysdate;

    }
    //--------------------------------------------------------------------------

  }

}
